# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import logging
import queue
import threading
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import datetime
from enum import Enum
from queue import Queue
from typing import List, Optional

import tzlocal
from apscheduler.schedulers.background import BackgroundScheduler
from requests import Session, HTTPError

from featureprobe.access_recorder import AccessRecorder
from featureprobe.context import Context
from featureprobe.event import Event, AccessEvent
from featureprobe.event_processor import EventProcessor


class EventAction:
    class Type(Enum):
        EVENT = 1
        FLUSH = 2
        SHUTDOWN = 3

    def __init__(self, _type: Type, event: Optional[Event]):
        self.type = _type
        self.event = event


class EventRepository:
    def __init__(self):
        self.events = []
        self.access = AccessRecorder()

    @classmethod
    def _clone(
            cls,
            events: List[Event],
            access: AccessRecorder) -> "EventRepository":
        repo = cls()
        repo.events = events.copy()
        repo.access = access.snapshot()
        return repo

    def to_dict(self) -> dict:
        return {
            'events': [e.to_dict() for e in self.events],
            'access': self.access.to_dict(),
        }

    def __bool__(self):
        return not self.is_empty()

    def is_empty(self):
        return len(self.events) == 0 and len(self.access.counters) == 0

    def add(self, event: Event):
        if isinstance(event, AccessEvent):
            self.access.add(event)

    def snapshot(self):
        return EventRepository._clone(self.events, self.access)

    def clear(self):
        self.events.clear()
        self.access.clear()


class DefaultEventProcessor(EventProcessor):
    _logger = logging.getLogger('FeatureProbe-Event')
    _EVENT_BATCH_HANDLE_SIZE = 50
    _CAPACITY = 10000

    _LOG_BUSY_EVENT = 'Event processing is busy, some will be dropped'

    def __init__(self, context: Context):
        self._closed = False
        self._events = Queue(maxsize=DefaultEventProcessor._CAPACITY)

        self._api_url = context.event_url
        self._session = Session()
        self._session.mount('http://', context.http_config.adapter)
        self._session.mount('https://', context.http_config.adapter)
        self._session.headers.update(context.headers)
        self._timeout = (
            context.http_config.conn_timeout,
            context.http_config.read_timeout)

        self._event_repository = EventRepository()
        self._handler_thread = threading.Thread(
            target=self._handle_event, args=(
                self._events, self._event_repository), daemon=True)
        self._handler_thread.start()

        self._executor = ThreadPoolExecutor(max_workers=5)
        self._scheduler = BackgroundScheduler(
            timezone=tzlocal.get_localzone(), logger=self._logger)
        self._scheduler.start()
        self._scheduler.add_job(self.flush,
                                trigger='interval',
                                seconds=5,
                                next_run_time=datetime.now())

    @classmethod
    def from_context(cls, context: Context) -> EventProcessor:
        return cls(context)

    def push(self, event: Event):
        if self._closed:
            return
        try:
            self._events.put_nowait(EventAction(
                EventAction.Type.EVENT, event))
        except queue.Full:
            DefaultEventProcessor._logger.warning(
                DefaultEventProcessor._LOG_BUSY_EVENT)

    def flush(self, block=False, timeout=None):
        if self._closed:
            return
        try:
            self._events.put(
                EventAction(EventAction.Type.FLUSH, None), block, timeout)
        except queue.Full:
            DefaultEventProcessor._logger.warning(
                DefaultEventProcessor._LOG_BUSY_EVENT)

    def shutdown(self):
        self._do_shutdown()

    def _handle_event(self, events: Queue, event_repo: EventRepository):
        actions = []
        while not self._closed or not events.empty():
            actions.clear()
            actions.append(events.get())
            with contextlib.suppress(queue.Empty):
                actions.extend(
                    events.get_nowait() for _ in range(
                        1, DefaultEventProcessor._EVENT_BATCH_HANDLE_SIZE))
            try:
                for action in actions:
                    if action.type == EventAction.Type.EVENT:
                        self._process_event(action.event, event_repo)
                    elif action.type == EventAction.Type.FLUSH:
                        self._process_flush(event_repo)
                    elif action.type == EventAction.Type.SHUTDOWN:
                        self._do_shutdown()
            except Exception as e:
                self._logger.error(
                    'FeatureProbe event handle error', exc_info=e)

    def _do_shutdown(self):
        if self._closed:
            return

        self.flush(block=True, timeout=2)
        self._closed = True
        self._handler_thread.join(2)
        self._scheduler.shutdown(wait=True)
        self._executor.shutdown(wait=True)

    def _process_event(self, event: Event, event_repo: EventRepository):
        event_repo.add(event)

    def _send_events(self, repositories: List[EventRepository]):
        repositories = [repo.to_dict() for repo in repositories]
        resp = self._session.post(
            self._api_url,
            json=repositories,
            timeout=self._timeout)
        # sourcery skip: replace-interpolation-with-fstring
        self._logger.debug('Http response: %s' % resp.text)
        try:
            resp.raise_for_status()
        except HTTPError as e:
            self._logger.error(
                'Unexpected error from event sender',
                exc_info=e)

    def _process_flush(self, event_repo: EventRepository):
        if not event_repo:
            return
        send_queue = [event_repo.snapshot()]
        self._executor.submit(self._send_events, send_queue)
        event_repo.clear()



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from featureprobe.context import Context
    from featureprobe.model.repository import Repository
    from featureprobe.model.toggle import Toggle


class DataRepository(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def from_context(cls, context: "Context") -> "DataRepository":
        pass

    @abstractmethod
    def refresh(self, repo: "Repository"):
        pass

    @abstractmethod
    def get_toggle(self, key: str) -> "Toggle":
        pass

    @abstractmethod
    def get_all_toggle(self) -> Dict[str, "Toggle"]:
        pass

    @abstractmethod
    def get_segment(self, key: str) -> "Toggle":
        pass

    @abstractmethod
    def get_all_segment(self) -> Dict[str, "Toggle"]:
        pass

    @property
    @abstractmethod
    def initialized(self) -> bool:
        pass

    @abstractmethod
    def close(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'FeatureProbe'
__license__ = 'Apache 2.0'


# PEP440 spec:
# [N!]N(.N)*[{a|b|rc}N][.postN][.devN]
__version__ = '1.1.1'


# --------------- API --------------- #

from featureprobe.model import *

from featureprobe.access_recorder import (
    AccessCounter,
    AccessRecorder,
)

from featureprobe.config import Config
from featureprobe.context import Context
from featureprobe.data_repository import DataRepository
from featureprobe.detail import Detail
from featureprobe.evaluation_result import EvaluationResult
from featureprobe.event import AccessEvent
from featureprobe.hit_result import HitResult
from featureprobe.http_config import HttpConfig
from featureprobe.client import Client
from featureprobe.user import User


__all__ = [
    # featureprobe.model

    'Condition',
    'ConditionType',
    'StringPredicate',
    'SegmentPredicate',
    'DatetimePredicate',
    'NumberPredicate',
    'SemverPredicate',
    'Distribution',
    'Repository',
    'Rule',
    'Segment',
    'SegmentRule',
    'Serve',
    'Split',
    'Toggle',

    # featureprobe

    'AccessCounter',
    'AccessRecorder',
    'Client',
    'Config',
    'Context',
    'DataRepository',
    'Detail',
    'EvaluationResult',
    'AccessEvent',
    'HitResult',
    'HttpConfig',
    'User',
]



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from datetime import datetime
from typing import TYPE_CHECKING

import tzlocal
from apscheduler.schedulers.background import BackgroundScheduler
from requests import Session

from featureprobe import Repository
from featureprobe.synchronizer import Synchronizer

if TYPE_CHECKING:
    from featureprobe.context import Context
    from featureprobe.data_repository import DataRepository


class PoolingSynchronizer(Synchronizer):
    __logger = logging.getLogger('FeatureProbe-Synchronizer')

    def __init__(self, context: "Context", data_repo: "DataRepository"):
        self._refresh_interval = context.refresh_interval
        self._api_url = context.synchronizer_url
        self._data_repo = data_repo

        self._session = Session()
        self._session.mount('http://', context.http_config.adapter)
        self._session.mount('https://', context.http_config.adapter)
        self._session.headers.update(context.headers)
        self._timeout = (
            context.http_config.conn_timeout,
            context.http_config.read_timeout)

        self._scheduler = None
        self._lock = threading.RLock()

    @classmethod
    def from_context(
            cls,
            context: "Context",
            data_repo: "DataRepository") -> "Synchronizer":
        return cls(context, data_repo)

    def sync(self):
        PoolingSynchronizer.__logger.info(
            'Starting FeatureProbe polling repository with interval %d ms'
            % (self._refresh_interval.total_seconds() * 1000))
        self._poll()
        with self._lock:
            self._scheduler = BackgroundScheduler(
                timezone=tzlocal.get_localzone(),
                logger=self.__logger)
            self._scheduler.start()
            self._scheduler.add_job(
                self._poll,
                trigger='interval',
                seconds=self._refresh_interval.total_seconds(),
                next_run_time=datetime.now())

    def close(self):
        PoolingSynchronizer.__logger.info(
            'Closing FeatureProbe PollingSynchronizer')
        with self._lock:
            self._scheduler.shutdown()
            del self._scheduler

    def _poll(self):
        try:
            resp = self._session.get(self._api_url, timeout=self._timeout)
            resp.raise_for_status()
            self.__logger.debug('Http response: %d' % resp.status_code)
            body = resp.json()
            # sourcery skip: replace-interpolation-with-fstring
            self.__logger.debug('Http response body: %s' % body)
            repo = Repository.from_json(body)
            self._data_repo.refresh(repo)
        except Exception as e:  # noqa
            self.__logger.error(
                'Unexpected error from polling processor',
                exc_info=e)



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta
from enum import Enum
from typing import Union

from featureprobe.default_event_processor import DefaultEventProcessor
from featureprobe.file_synchronizer import FileSynchronizer
from featureprobe.http_config import HttpConfig
from featureprobe.internal.defaultable import defaultable
from featureprobe.memory_data_repository import MemoryDataRepository
from featureprobe.pooling_synchronizer import PoolingSynchronizer


class SyncMode(str, Enum):
    def __new__(cls, value, synchronizer_creator):
        if isinstance(value, cls):
            return value
        if synchronizer_creator is None:
            return cls._value2member_map_[value]

        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.synchronizer_creator = synchronizer_creator
        return obj

    POOLING = 'pooling', PoolingSynchronizer.from_context
    FILE = 'file', FileSynchronizer.from_context


@defaultable
class Config:

    def __init__(self,
                 location: str = None,
                 sync_mode: Union[str, SyncMode] = SyncMode.POOLING,
                 synchronizer_url: str = None,
                 event_url: str = None,
                 remote_uri: str = 'http://127.0.0.1:4007',
                 http_config: HttpConfig = HttpConfig(),
                 refresh_interval: Union[timedelta, float] = timedelta(seconds=5),
                 ):
        self._location = location
        self._synchronizer_creator = SyncMode(sync_mode).synchronizer_creator
        self._data_repository_creator = MemoryDataRepository.from_context
        self._event_processor_creator = DefaultEventProcessor.from_context
        self._synchronizer_url = synchronizer_url
        self._event_url = event_url
        self._remote_uri = remote_uri
        self._http_config = http_config or HttpConfig()
        self._refresh_interval = refresh_interval \
            if isinstance(refresh_interval, timedelta) \
            else timedelta(seconds=refresh_interval)

    @property
    def location(self):
        return self._location

    @property
    def synchronizer_creator(self):
        return self._synchronizer_creator

    @property
    def data_repository_creator(self):
        return self._data_repository_creator

    @property
    def event_processor_creator(self):
        return self._event_processor_creator

    @property
    def synchronizer_url(self):
        return self._synchronizer_url

    @property
    def event_url(self):
        return self._event_url

    @property
    def remote_uri(self):
        return self._remote_uri

    @property
    def http_config(self):
        return self._http_config

    @property
    def refresh_interval(self):
        return self._refresh_interval



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Any

from featureprobe.config import Config
from featureprobe.context import Context
from featureprobe.detail import Detail
from featureprobe.event import AccessEvent
from featureprobe.internal.empty_str import empty_str
from featureprobe.user import User


class Client:
    """A client for the FeatureProbe API. Client instances are thread-safe.

    Applications should instantiate a single :obj:`~featureprobe.Client` for the lifetime of their application.
    """

    __logger = logging.getLogger('FeatureProbe')

    def __init__(self, server_sdk_key: str, config: Config = Config()):
        """Creates a new client instance that connects to FeatureProbe with the default configuration.

        :param server_sdk_key: Server SDK Key for your FeatureProbe environment.
        :param config: (optional) The configuration control FeatureProbe client behavior.
                        Leaving this argument unfilled will use the default configuration.
        """
        if empty_str(server_sdk_key):
            raise ValueError('sdk key must not be blank')
        context = Context(server_sdk_key, config)
        self._event_processor = config.event_processor_creator(context)
        self._data_repo = config.data_repository_creator(context)
        self._synchronizer = config.synchronizer_creator(
            context, self._data_repo)
        self._synchronizer.sync()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Alias for :func:`~featureprobe.Client.close`

        Usage::

          >>> import featureprobe as fp
          >>> with fp.Client('key_000') as client:
          >>>     ...
          >>> # client will be closed here
        """
        self.close()

    def flush(self):
        """Manually push events"""
        self._event_processor.flush()

    def close(self):
        """Safely shut down FeatureProbe client instance"""
        Client.__logger.info('Closing FeatureProbe Client')
        self._event_processor.shutdown()
        self._synchronizer.close()
        self._data_repo.close()

    def value(self, toggle_key: str, user: User, default) -> Any:
        """Gets the evaluated value of a toggle.

        :param toggle_key: The key of toggle in this environment.
        :param user: :obj:`~featureprobe.User` to be evaluated.
        :param default: The default value to be returned.
        :returns: Dependents on the toggle's type.
        """
        toggle = self._data_repo.get_toggle(toggle_key)
        segments = self._data_repo.get_all_segment()
        if not toggle:
            return default

        eval_result = toggle.eval(user, segments, default)
        access_event = AccessEvent(timestamp=int(time.time() * 1000),
                                   user=user,
                                   key=toggle_key,
                                   value=str(eval_result.value),
                                   version=eval_result.version,
                                   index=eval_result.variation_index)
        self._event_processor.push(access_event)
        return eval_result.value

    def value_detail(self, toggle_key: str, user: User, default) -> Detail:
        """Gets the detailed evaluated results of a toggle.

        :param toggle_key: The key of toggle in this environment.
        :param user: :obj:`~featureprobe.User` to be evaluated.
        :param default: The default value to be returned.
        :returns: :obj:`~featureprobe.Detail` contains the `value`, `rule_index`, `version`, and `reason`
                  of this evaluation.
        """

        if not self._data_repo.initialized:
            return Detail(
                value=default,
                reason='FeatureProbe repository uninitialized')

        toggle = self._data_repo.get_toggle(toggle_key)
        segments = self._data_repo.get_all_segment()

        if toggle is None:
            return Detail(value=default, reason='Toggle not exist')

        eval_result = toggle.eval(user, segments, default)
        detail = Detail(value=eval_result.value,
                        reason=eval_result.reason,
                        rule_index=eval_result.rule_index,
                        version=eval_result.version)
        access_event = AccessEvent(timestamp=int(time.time() * 1000),
                                   user=user,
                                   key=toggle_key,
                                   value=eval_result.value,
                                   version=eval_result.version,
                                   index=eval_result.variation_index)
        self._event_processor.push(access_event)
        return detail



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from featureprobe.user import User


class Event:
    def __init__(self, created_time: int, user: "User"):
        self._created_time = created_time
        self._user = user

    def to_dict(self) -> dict:
        return {
            'createdTime': self._created_time,
            'user': self._user.to_dict(),
        }

    @property
    def created_time(self) -> int:
        return self._created_time

    @property
    def user(self) -> "User":
        return self._user


class AccessEvent(Event):
    def __init__(
            self,
            timestamp: int,
            user: "User",
            key: str,
            value: str,
            version: int,
            index: int):
        super().__init__(timestamp, user)
        self._key = key
        self._value = value
        self._version = version
        self._index = index

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value

    @property
    def version(self):
        return self._version

    @property
    def index(self):
        return self._index



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from featureprobe.config import Config


class Context:
    _GET_REPOSITORY_DATA_API = '/api/server-sdk/toggles'
    _POST_EVENTS_DATA_API = '/api/events'

    def __init__(self, sdk_key: str, config: "Config"):
        from featureprobe import __version__
        self._synchronizer_url = config.synchronizer_url or (
            config.remote_uri + self._GET_REPOSITORY_DATA_API)
        self._event_url = config.event_url or (
            config.remote_uri + self._POST_EVENTS_DATA_API)
        self._sdk_key = sdk_key
        self._refresh_interval = config.refresh_interval
        self._location = config.location
        self._http_config = config.http_config
        self._headers = {
            'Authorization': sdk_key,
            'user-agent': 'Python/' + str(__version__),
        }

    @property
    def synchronizer_url(self):
        return self._synchronizer_url

    @property
    def event_url(self):
        return self._event_url

    @property
    def sdk_key(self):
        return self._sdk_key

    @property
    def refresh_interval(self):
        return self._refresh_interval

    @property
    def location(self):
        return self._location

    @property
    def http_config(self):
        return self._http_config

    @property
    def headers(self):
        return self._headers



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from featureprobe.context import Context
    from featureprobe.event import Event


class EventProcessor(ABC):

    @classmethod
    @abstractmethod
    def from_context(cls, context: "Context") -> "EventProcessor":
        pass

    @abstractmethod
    def push(self, event: "Event"):
        pass

    @abstractmethod
    def flush(self):
        pass

    @abstractmethod
    def shutdown(self):
        pass



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta
from typing import Union

from requests.adapters import HTTPAdapter

from featureprobe.internal.defaultable import defaultable


@defaultable
class HttpConfig:
    def __init__(self,
                 pool_connections: int = 5,
                 pool_maxsize: int = 10,
                 conn_timeout: Union[timedelta, float] = timedelta(seconds=3),
                 read_timeout: Union[timedelta, float] = timedelta(seconds=3)):
        self.conn_timeout = conn_timeout.total_seconds() \
            if isinstance(conn_timeout, timedelta) \
            else conn_timeout
        self.read_timeout = read_timeout.total_seconds() \
            if isinstance(read_timeout, timedelta) \
            else read_timeout
        self.adapter = HTTPAdapter(
            pool_maxsize=pool_maxsize,
            pool_connections=pool_connections,
            max_retries=0,
        )



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Dict, Optional, TYPE_CHECKING

from featureprobe.data_repository import DataRepository
from featureprobe.model.repository import Repository

if TYPE_CHECKING:
    from featureprobe.context import Context
    from featureprobe.model.segment import Segment
    from featureprobe.model.toggle import Toggle


class MemoryDataRepository(DataRepository):
    def __init__(self,
                 data: Repository,
                 initialized: bool,
                 updated_timestamp: int):
        self._data = data
        self._initialized = initialized
        self._updated_timestamp = updated_timestamp

    @classmethod
    def from_context(cls, context: "Context") -> DataRepository:
        return cls(data=None, initialized=False, updated_timestamp=0)  # noqa

    def refresh(self, repo: Repository):
        if repo is not None \
                and repo.toggles is not None \
                and repo.segments is not None:
            self._data = Repository(repo.toggles.copy(), repo.segments.copy())
            self._initialized = True
            self._updated_timestamp = int(time.time() * 1000)

    def get_toggle(self, key: str) -> Optional["Toggle"]:
        return self._data.toggles.get(key) if self._initialized else None

    def get_all_toggle(self) -> Dict[str, "Toggle"]:
        return self._data.toggles if self._initialized else {}

    def get_segment(self, key: str) -> Optional["Segment"]:
        return self._data.segments.get(key) if self._initialized else None

    def get_all_segment(self) -> Dict[str, "Toggle"]:
        return self._data.segments if self._initialized else {}

    def initialized(self) -> bool:
        return self._initialized

    def close(self):
        self._data = None



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from typing import TYPE_CHECKING

from featureprobe.model.repository import Repository
from featureprobe.synchronizer import Synchronizer

if TYPE_CHECKING:
    from featureprobe.context import Context
    from featureprobe.data_repository import DataRepository


class FileSynchronizer(Synchronizer):
    _logger = logging.getLogger('FeatureProbe-Synchronizer')

    def __init__(self,
                 data_repository: "DataRepository",
                 location: str):
        self._data_repository = data_repository
        self._location = location

    @classmethod
    def from_context(
            cls,
            context: "Context",
            data_repo: "DataRepository") -> "Synchronizer":
        return cls(data_repo, context.location)

    def sync(self):
        try:
            with open(self._location, 'r', encoding='utf-8') as f:
                repo = Repository.from_json(json.load(f))
                self._data_repository.refresh(repo)
        except FileNotFoundError:
            # sourcery skip: replace-interpolation-with-fstring
            self._logger.error(
                'repository file resource not found in path: %s' %
                self._location)

    def close(self):
        return



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Dict


class User:
    """A collection of attributes that can affect toggle evaluation.

    Usually corresponding to a user of your application.
    """

    def __init__(self, attrs: Dict[str, str] = None,
                 stable_rollout_key: str = None):
        """Creates a new FeatureProbe User.

        :param attrs: (optional) The initialize attribute for a user.
        :param stable_rollout_key: User unique ID for percentage rollout.
        """
        if stable_rollout_key is not None:
            self._key = stable_rollout_key
        else:
            self._key = str(int(time.time() * 10**6))
        self._attrs = attrs or {}

    def __setitem__(self, key: str, value: str):
        """Alias for :func:`~featureprobe.User.with_attr`.

        Usage::

          >>> import featureprobe as fp
          >>> user = fp.User('unique id')
          >>> user.with_attr('key1', 'value1')
          >>> user['key2'] = 'value2'
        """
        self._attrs[key] = value

    def __getitem__(self, attr: str):
        """Gets the value of specified attribute.

        :param attr: Attribute name / key.
        :returns: str or None

        Usage::

          >>> import featureprobe as fp
          >>> user = fp.User('unique id', { 'key': 'value' })
          >>> user['key']  # 'value'
        """
        try:
            return self._attrs[attr]
        except KeyError:
            return None

    def __delitem__(self, key: str):
        self._attrs.pop(key, None)

    def stable_rollout(self, key):
        self._key = key
        return self

    def to_dict(self) -> dict:
        return {
            'key': self._key,
            'attrs': self._attrs,
        }

    @property
    def key(self) -> str:
        """Gets FeatureProbe User unique identifier"""
        return self._key

    @property
    def attrs(self) -> Dict[str, str]:
        """Gets all attributes of a FeatureProbe User"""
        return self._attrs

    @attrs.setter
    def attrs(self, attrs: Dict[str, str]):
        """Sets (replace the original attributes) multiple attributes to a FeatureProbe User"""
        self._attrs = attrs

    def with_attr(self, key: str, value: str) -> "User":
        """Adds an attribute to the user.

        :param key: Attribute key / name.
        :param value: Attribute value.
        :returns: User

        Usage::

          >>> import featureprobe as fp
          >>> user = fp.User('unique id').with_attr('key1', 'value1').with_attr('key2', 'value2')
        """
        self._attrs[key] = value
        return self

    def has_attr(self, attr: str) -> bool:
        """Checks if an attribute exists.

        :param attr: Attribute name / key.
        :returns: bool
        """
        return attr in self._attrs



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from featureprobe.event import AccessEvent


class AccessCounter:
    def __init__(self, value: str, version: int, index: int):
        self._VALUE = value
        self._VERSION = version
        self._INDEX = index
        self._count = 1

    def __str__(self):
        return "AccessCounter(value='%s', version=%d, index=%d, count=%d)" \
               % (self._VALUE, self._VERSION, self._INDEX, self._count)

    def to_dict(self) -> dict:
        return {
            'value': self._VALUE,
            'version': self._VERSION,
            'index': self._INDEX,
            'count': self._count,
        }

    @property
    def value(self):
        return self._VALUE

    @property
    def version(self):
        return self._VERSION

    @property
    def index(self):
        return self._INDEX

    @property
    def count(self):
        return self._count

    def increment(self):
        self._count += 1

    def is_group(self, event: "AccessEvent"):
        return self._VALUE == event.value \
            and self._VERSION == event.version \
            and self._INDEX == event.index


class AccessRecorder:
    def __init__(self):
        self._counters = {}  # Dict[str, List[AccessCounter]]
        self._start_time = 0
        self._end_time = 0

    def to_dict(self) -> dict:
        return {
            'counters': {k: [ac.to_dict() for ac in v] for k, v in self._counters.items()},
            'startTime': self._start_time,
            'endTime': self._end_time,
        }

    @property
    def counters(self):
        return self._counters

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    def add(self, _event: "AccessEvent"):  # sourcery skip: use-named-expression
        if not self._counters:
            self._start_time = int(time.time() * 1000)
        counters = self._counters.get(_event.key)
        if counters:
            for counter in counters:
                if counter.is_group(_event):
                    counter.increment()
                    return
            counters.append(
                AccessCounter(
                    _event.value,
                    _event.version,
                    _event.index))
        else:
            groups = [
                AccessCounter(
                    _event.value,
                    _event.version,
                    _event.index)]
            self._counters[_event.key] = groups

    def snapshot(self):
        _snapshot = copy.deepcopy(self)
        _snapshot._end_time = int(time.time() * 1000)
        return _snapshot

    def clear(self):
        self._counters = {}



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional


class Detail:
    def __init__(self, value=None,
                 rule_index: Optional[int] = None,
                 version: Optional[int] = None,
                 reason: Optional[str] = None):
        self.value = value
        self.rule_index = rule_index
        self.version = version
        self.reason = reason

    def __str__(self):
        return "FPDetail{value=%s, rule_index=%s, version=%s, reason='%s'}" % \
               (self.value, self.rule_index, self.version, self.reason)
